package kr.co.iya.base.advice;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.time.Duration;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.RestControllerAdvice;
import org.springframework.web.context.request.WebRequest;

import com.fasterxml.jackson.core.type.TypeReference;

import kr.co.iya.base.exception.BusinessException;
import kr.co.iya.base.exception.SystemException;
import kr.co.iya.base.support.AbstractMeta;
import kr.co.iya.base.support.util.SystemUtilPack.SystemUtil;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@RestControllerAdvice
public class ExceptionAdvice extends AbstractMeta {

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	private final static String DEFAULT_MESSAGE = "시스템 이용에 불편을 드려서 죄송합니다.";
	
	@Value("${spring.application.name}")
	private String applicationName;

	@Value("${spring.profiles.active:local}")
	private String springProfilesActive;
	
	@Value("${project.name:}")
	private String projectName;	
	
	@FunctionalInterface
	public interface ExceptionMessageLambda {public abstract String getMessage();}

	public static String message(String message) {return makeMessage(message,null);}
	public static String message(String message, String arg1) {return makeMessage(message,new Object[] {arg1});}
	public static String message(String message, Object[] args) {return makeMessage(message,args);}

	public static String getTrace(Throwable e) {
		try {
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			PrintStream stream = new PrintStream(out);
			e.printStackTrace(stream);
			return out.toString();
		}catch(Exception ex) {
			return ex.getMessage();
		}
	}

	public static Duration getExceptionTraceDuration() {
		return Duration.ofMinutes(10);
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		
	private static String makeMessage(String message, Object[] args) {		
		String realMessage=MetaUtil.getMessageUtil().getMessage(message==null?"":message,args);
		
		boolean isCode=true;
		if(isCode && message==null) {isCode=false;}
		if(isCode && message.equals(realMessage)) {isCode=false;}
		
		Map<String,Object> map = new HashMap<>();
		map.put("messageCode",message);
		map.put("message",realMessage);
		
		return MetaUtil.getJsonUtil().toJson(map);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	@ExceptionHandler({BusinessException.class,})
	protected ResponseEntity<Object> businessExceptionHandler(BusinessException ex, WebRequest request){
		Map<String,Object> body = new HashMap<>();
		body.put("httpStatus",ex.getHttpStatus().value());
		body.putAll(this.parseMessage(ex,request));

		return new ResponseEntity<>(body,ex.getHttpStatus());
	}
	
	@ExceptionHandler({SystemException.class})
	protected ResponseEntity<Object> systemExceptionHandler(SystemException ex, WebRequest request){
		Map<String,Object> body = new HashMap<>();
		body.put("httpStatus",ex.getHttpStatus().value());
		body.putAll(this.parseMessage(ex,request));

		return new ResponseEntity<>(body,ex.getHttpStatus());
	}	

	@ExceptionHandler(IllegalArgumentException.class)
	protected ResponseEntity<Object> illegalArgumentExceptionHandler(IllegalArgumentException ex, WebRequest request){
		Map<String,Object> body = new HashMap<>();
		body.put("httpStatus",HttpStatus.CONFLICT); //409
		body.putAll(this.parseMessage(ex,request));		

		return new ResponseEntity<>(body,HttpStatus.INTERNAL_SERVER_ERROR);
	}	
	
	@ExceptionHandler(Exception.class)
	protected ResponseEntity<Object> exceptionHandler(Exception ex, WebRequest request){
		Map<String,Object> body = new HashMap<>();
		body.put("httpStatus",HttpStatus.INTERNAL_SERVER_ERROR); //500
		body.putAll(this.parseMessage(ex,request));
		//body.put("trace", traceStr);
		
		return new ResponseEntity<>(body,HttpStatus.INTERNAL_SERVER_ERROR);
	}	
	
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	private Map<String,Object> parseMessage(Throwable ex, WebRequest request){
		String message = ex.getMessage();
		String exceptionName = ex.getClass().getName();
		
		Map<String,Object> map = new HashMap<>();
		
		if(!this.springProfilesActive.equals(SystemUtil.PROFILE_PROD)) {		
			map.put("exception",exceptionName);
		}
		map.put("propagation",this.projectName);
		map.put("date",new Date());
		map.put("messageCode","");
		map.put("message",message);

		StringBuilder traceId = new StringBuilder();
		traceId.append(new SimpleDateFormat("yyyyMMddHHmmssSSS").format(System.currentTimeMillis())).append(":"); //when
		traceId.append(MetaUtil.getSystemUtil().getServerInstanceCode()).append(":"); //where[base64(ip)+port]
		traceId.append(this.applicationName).append(":"); //application
		traceId.append(exceptionName.substring(exceptionName.lastIndexOf(".")+1)); //exption
				
		log.error("\ntraceId:{}\n{}",traceId.toString(),getTrace(ex));

		//empty인경우 디폴트메세지 처리
		//kr.co.iya 패키지 하부의 Exception이 아닌 runtimeException이 throw된 경우는 개발자가 의도적으로 메세지 처리를 하지 않은 경우로 디폴트메세지를 사용한다.
		//kr.co.iya 패키지 하부의 Exception을 사용한 경우는 메세지가 구성된 Exception으로 메세지를 추출하여 사용한다.
		boolean isDefault=false;
		if(!isDefault && StringUtils.isBlank(message)) {isDefault=true;}
		if(!isDefault && ex.getClass().getCanonicalName().indexOf("kr.co.iya")==-1) {isDefault=true;}
		if(isDefault) {
				log.debug(">> Change Default Message:[{}]->[{}]",message,DEFAULT_MESSAGE);
				map.put("message",DEFAULT_MESSAGE);
		}

		//오류가 타 시스템으로 부터 전파경우 메세지 처리
		if(MetaUtil.getJsonUtil().isJsonObject(map.get("message"))) {
			try {
				Map<String,Object> refMap = MetaUtil.getObjectMapperUtil().readValue(String.valueOf(map.get("message")), new TypeReference<Map<String,Object>>(){});
				
				if(refMap.containsKey("propagation")) {map.put("propagation",refMap.get("propagation")+">"+map.get("propagation"));}
				if(refMap.containsKey("messageCode")) {map.put("messageCode",refMap.get("messageCode"));}
				if(refMap.containsKey("message")) {map.put("message",refMap.get("message"));}
				
			}catch(Exception e){
				log.debug(">> Change Default Message:[{}]->[{}]",map.get("message"),DEFAULT_MESSAGE);
				map.put("message",DEFAULT_MESSAGE);
				log.debug(">> 에러메세지 처리시 오류 발생, 처리대상 메세지:{}",map.get("message"));
				log.debug(">> 디폴드 메세지를 사용함 :{}",DEFAULT_MESSAGE);
			}			
		}

		//운영환경이 아닌 경우, 상세 에러메세지를 조회할 수 있는 traceId를 붙인다
		if(!this.springProfilesActive.equals(SystemUtil.PROFILE_PROD)) {
			StringBuilder builder = new StringBuilder();
			builder.append(map.get("message"));
			builder.append("(traceId:").append(traceId).append(")");
			map.put("message",builder.toString());
		}

		return map;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}