package kr.co.iya.base.config;

import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.condition.ConditionalOnExpression;
import org.springframework.boot.autoconfigure.condition.ConditionalOnWebApplication;
import org.springframework.boot.web.error.ErrorAttributeOptions;
import org.springframework.boot.web.servlet.error.DefaultErrorAttributes;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.http.HttpStatus;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.web.context.request.RequestAttributes;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.config.annotation.CorsRegistry;
import org.springframework.web.servlet.config.annotation.InterceptorRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurer;
import org.springframework.web.servlet.i18n.LocaleChangeInterceptor;
import org.springframework.web.servlet.i18n.SessionLocaleResolver;

import kr.co.iya.base.advice.PageAdvice;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@Configuration
@ConditionalOnExpression("T(kr.co.iya.base.WebMvcConfig).isEnabled()")
@ConditionalOnWebApplication
public class WebMvcConfig implements WebMvcConfigurer {

	@Value("${project.name:}")
	private String projectName;	

	@Value("${interceptor.enabled:true}")
	private boolean isEnable_interceptor;

	@Value("${interceptor.locale.enabled:true}")
	private boolean isEnable_localeChangeInterceptor;

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	@Override	
	public void addCorsMappings(CorsRegistry registry) {
		registry.addMapping("/**")
			.maxAge(1800L) // Access-Control-Max-Age : 클라이언트에서 preFlight의 요청 결과를 저장할 시간(초) 지정
			.allowedOriginPatterns("*") // Access-Control-Allow-Origin : 요청을 보내는 페이지의 출처 (*, 도메인)
			.allowedHeaders("*") // Access-Control-Allow-Headers : 요청을 허용하는 헤더
			.allowedMethods("*") // Access-Control-Allow-Methods : 요청을 허용하는 메소드 (Default : GET, POST, HEAD)
			.allowCredentials(true) // Access-Control-Allow-Credentials : 쿠키를 요청에 포함 (요청시 withCredentials:true)
			.exposedHeaders(PageAdvice.getPagingKey()); // JavaScript에서 참조하기 위한 추가, CORS의 경우 기본적으로 화면(JavaScript)에서 response header 값을 읽지 못함.
	}
	
    @Override
    public void addInterceptors(InterceptorRegistry registry) {    	
    	if(!isEnable_interceptor) {return;}

    	int order=0;
        if(isEnable_localeChangeInterceptor) {registry.addInterceptor(localeChangeInterceptor()).order(++order);}
    }

    //@Override
    //public void addArgumentResolvers(List<HandlerMethodArgumentResolver> argumentResolvers) {
    //	 아규먼트리졸버(Argument Resolver)는 사용자가 컨트롤러의 메서드 인자값으로 임의의 값을 전달하려할 때 사용
    //   argumentResolvers.add(deviceHandlerMethodArgumentResolver());
    //}
    
    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    @Bean
    DefaultErrorAttributes errorAttributes() {		
		return new DefaultErrorAttributes() {
			@Override
			public Map<String,Object> getErrorAttributes(WebRequest webRequest, ErrorAttributeOptions options) {
				Integer status = (Integer) webRequest.getAttribute("javax.servlet.error.status_code", RequestAttributes.SCOPE_REQUEST);				
				if(status!=HttpStatus.NOT_FOUND.value()) {return super.getErrorAttributes(webRequest, options);}
				
				Map<String, Object> errorAttributes = new HashMap<>();				
				errorAttributes.put("httpStatus", HttpStatus.NOT_FOUND.value());
				errorAttributes.put("messageCode", "404");				
				errorAttributes.put("message", "Not Found");
				errorAttributes.put("exception", "");
				errorAttributes.put("propagation", projectName);
	
				return errorAttributes;
			}
		};
	}

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Bean
    @ConditionalOnExpression("'${interceptor.enabled:true}'.equals('true') && '${interceptor.locale.enabled:true}'.equals('true')")
    HandlerInterceptor localeChangeInterceptor() {
    	LocaleChangeInterceptor interceptor = new LocaleChangeInterceptor();
        interceptor.setParamName("lang");
    	log.info(">> interceptor-localeChangeInterceptor [paramName:{}]",interceptor.getParamName());
        return interceptor;
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    //@Bean
    //public DeviceHandlerMethodArgumentResolver deviceHandlerMethodArgumentResolver() {
    //   return new DeviceHandlerMethodArgumentResolver();
    //}

    @Bean
    SessionLocaleResolver localeResolver() {
    	SessionLocaleResolver resolver = new SessionLocaleResolver();
    	resolver.setDefaultLocale(Locale.KOREA);
    	log.info(">> localeResolver.locale:{}",Locale.KOREA);
    	return resolver;
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Bean
    HttpMessageConverter<String> messageConverter() {
        return new StringHttpMessageConverter(Charset.forName("UTF-8"));
    }

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//@Bean
	//public HttpSessionListener getHttpSessionListener() {
	//    return new HttpSessionListener() {
	//    	
	//        @Override
	//        public void sessionCreated(HttpSessionEvent se) {
	//        	HttpSession s = se.getSession();
	//    		log.debug(">> 세션.생성[id:{},creationTime:{},maxInactiveInterval:{}]",s.getId(),s.getCreationTime());
	//        }
	//
	//        @Override
	//        public void sessionDestroyed(HttpSessionEvent se) {
	//        	HttpSession s = se.getSession();
	//    		log.debug(">> 세션.종료[id:{},creationTime:{},maxInactiveInterval:{}]",s.getId(),s.getCreationTime());
	//        }
	//    };
	//}


	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//addCorsMappings 메소드 적용으로 주석처리
	//	@Bean
	//	public FilterRegistrationBean<Filter> filterRegistrationBean() { 
	//		CorsConfiguration config = new CorsConfiguration();
	//		config.setMaxAge(1800L); // Access-Control-Max-Age : 클라이언트에서 preFlight의 요청 결과를 저장할 시간(초) 지정 (default:1800L)
	//		config.addAllowedOrigin("*"); // Access-Control-Allow-Orgin : 요청을 보내는 페이지의 출처 (*, 도메인)
	//		config.addAllowedHeader("*"); // Access-Control-Allow-Headers : 요청을 허용하는 헤더
	//		config.addAllowedMethod("*"); // Access-Control-Allow-Methods : 요청을 허용하는 메소드 (Default : GET, POST, HEAD)
	//		config.setAllowCredentials(true); // Access-Control-Allow-Credentials : 쿠키를 요청에 포함 (요청시 withCredentials:true)
	//		config.addExposedHeader(PageAdvice.getPagingKey()); // JavaScript에서 참조하기 위한 추가, CORS의 경우 기본적으로 화면(JavaScript)에서 response header 값을 읽지 못함.
	//		
	//		UrlBasedCorsConfigurationSource source = new UrlBasedCorsConfigurationSource();
	//		source.registerCorsConfiguration("/**", config);
	//		
	//		Filter filter = new BaseCorsFilter(source);
	//		
	//		FilterRegistrationBean<Filter> bean = new FilterRegistrationBean<>(filter);
	//		bean.setOrder(0);
	//		
	//		return bean;
	//	} 
    
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}